#include "assert.h"
#include "ctype.h"
#include "stdio.h"
#include "zlib.h"
#include "Error.h"


#include "GenomeSequence.h"

#include <fstream>

//
// Map ASCII values to a 2 (or 3) bit encoding for the base pair value
//  class 0 -> 'A' (Adenine - 0x41 and 0x61)
//  class 1 -> 'C' (Cytosine - 0x43 and 0x63)
//  class 2 -> 'G' (Guanine - 0x47 and 0x67)
//  class 3 -> 'T' (Thymine - 0x54 and 0x74)
//  class 4 -> 'N' (Unknown - read error or incomplete data - 0x4E and 0x6E)
//  class 5 -> not a valid DNA base pair character
//
// Note: The +1 array size is for the terminating NUL character
unsigned char GenomeSequence::base2int[256+1] =
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0x00-0x0F
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0x10-0x1F
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0x20-0x2F
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0x30-0x3F
"\005\000\005\001\005\005\005\002\005\005\005\005\005\005\004\005"  // 0x40-0x4F
"\005\005\005\005\003\005\005\005\005\005\005\005\005\005\005\005"  // 0x50-0x5F
"\005\000\005\001\005\005\005\002\005\005\005\005\005\005\004\005"  // 0x60-0x6F
"\005\005\005\005\003\005\005\005\005\005\005\005\005\005\005\005"  // 0x70-0x7F
// not used, but included for completeness:
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0x80-0x8F
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0x90-0x9F
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0xA0-0xAF
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0xB0-0xBF
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0xC0-0xCF
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0xD0-0xDF
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0xE0-0xEF
"\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005"  // 0xF0-0xFF
;

const char GenomeSequence::int2base[] = "ACGTNMXXXXXXXXXX";

unsigned char GenomeSequence::base2complement[256+1 /* for NUL char */] =
"NNNNNNNNNNNNNNNN"  // 0x00-0x0F
"NNNNNNNNNNNNNNNN"  // 0x10-0x1F
"NNNNNNNNNNNNNNNN"  // 0x20-0x2F
"NNNNNNNNNNNNNNNN"  // 0x30-0x3F
"NTNGNNNCNNNNNNNN"  // 0x40-0x4F
"NNNNANNNNNNNNNNN"  // 0x50-0x5F
"NTNGNNNCNNNNNNNN"  // 0x60-0x6F
"NNNNANNNNNNNNNNN"  // 0x70-0x7F
// not used, but included for completeness:
"NNNNNNNNNNNNNNNN"  // 0x80-0x8F
"NNNNNNNNNNNNNNNN"  // 0x90-0x9F
"NNNNNNNNNNNNNNNN"  // 0xA0-0xAF
"NNNNNNNNNNNNNNNN"  // 0xB0-0xBF
"NNNNNNNNNNNNNNNN"  // 0xC0-0xCF
"NNNNNNNNNNNNNNNN"  // 0xD0-0xDF
"NNNNNNNNNNNNNNNN"  // 0xE0-0xEF
"NNNNNNNNNNNNNNNN"  // 0xF0-0xFF
;

//
// given a read in a string, pack it into the vector of
// bytes coded as two bases per byte.
//
// The goal is to allow us to more rapidly compare against
// the genome, which is itself packed 2 bases per byte.
//
// Unfortunately, the match position may be odd or even, 
// so as a result, we also need to be able to prepad
// with 'N' bases to get the byte alignment the same, so
// padWithNCount may be a positive number indicating how
// many N bases to prepend.
//
void PackedRead::set(const char *rhs, int padWithNCount)
{
	clear();

	// pad this packed read with 'N' bases two at a time
	while(padWithNCount>1) {
		packedBases.push_back(
			GenomeSequence::base2int[(int) 'N'] << 4 |
			GenomeSequence::base2int[(int) 'N']
		);
		padWithNCount -= 2;
		length+=2;
	}

	// when we have only one base, pack one 'N' base with
	// the first base in rhs if there is one.
	if(padWithNCount) {
		// NB: *rhs could be NUL, which is ok here - just keep
		// the length straight.
		packedBases.push_back(
			GenomeSequence::base2int[(int) *rhs] << 4 |
			GenomeSequence::base2int[(int) 'N']
		);
		// two cases - have characters in rhs or we don't:
		if(*rhs) {
			length+=2;  // pad byte plus one byte from rhs
			rhs++;
		} else {
			length++;
		}
		padWithNCount--;    // should now be zero, so superfluous.
		assert(padWithNCount==0);
	}

	// pad pairs of bases from rhs, two at a time:
	while(*rhs && *(rhs+1)) {
		packedBases.push_back(
			GenomeSequence::base2int[(int) *(rhs+1)] << 4 |
			GenomeSequence::base2int[(int) *(rhs+0)]
		);
		rhs+=2;
		length+=2;
	}

	// if there is an odd base left at the end, put it
	// in a byte all its own (low 4 bits == 0):
	if(*rhs) {
		packedBases.push_back(
			GenomeSequence::base2int[(int) *(rhs+0)]
		);
		length++;
	}
	return;
}

String GenomeSequence::IntegerToSeq (unsigned int n, unsigned int wordsize)
{
	String sequence ("");
	for (unsigned int i = 0; i < wordsize; i ++)
		sequence += "N";

	unsigned int clearHigherBits = ~(3U << (wordsize<<1));	// XXX refactor - this appears several places

	if (n > clearHigherBits)
		error ("%d needs to be a non-negative integer < clearHigherBits\n", n);

	for (unsigned int i = 0; i < wordsize; i++)
	{
		sequence[wordsize-1-i] = int2base[n & 3];
		n >>= 2;
	}
	return sequence;
}



GenomeSequence::GenomeSequence()
{
}

GenomeSequence::~GenomeSequence()
{
	// free up resources:
	umfaFile.close();
}

//
// mapped open.
//
// if the file exists, map in into memory, and fill in a few useful
// fields.
//
// isColorSpace is used only when the baseFilename is empty, so
// that we can choose an appropriate default reference file.
//

int GenomeSequence::open(bool isColorSpace)
{
	int rc;

	if(baseFilename=="") {
		String name;
		if(isColorSpace) {
			name = "/usr/cluster/share/karma/NCBI36_5000CS.fa";
		} else {
			name = "/usr/cluster/share/karma/NCBI36_5000b.fa";
		}
		setFastaName(name);
	}
	rc = genomeSequenceArray::open(baseFilename + ".umfa");
	if (rc) {
		fprintf(stderr, "GenomeSequence::open: failed to open file %s.umfa\n", (const char *) baseFilename);
		return -1;
	}

	prefetch();

	return 0;
}

void GenomeSequence::sanityCheck(MemoryMap &fasta)
{
	unsigned int i;

	unsigned int genomeIndex = 0;
	for(i=0;i<fasta.length(); i++) {
		switch(fasta[i]) {
			case '>':
				while(fasta[i]!='\n' && fasta[i]!='\r') i++;
				break;
			case '\n':
			case '\r':
				break;
			default:
				assert(base2int[(int) (*this)[genomeIndex]] == base2int[(int) fasta[i]]);
#if 0
				if(base2int[(*this)[genomeIndex]] != base2int[(int) fasta[i]]) {
					int bp1 = base2int[(*this)[genomeIndex]];
					int bp2 = base2int[fasta[i]];
					printf("failing at genome index = %u, fasta index = %u.\n",
						genomeIndex,i);
				}
#endif
				genomeIndex++;
				break;
		}
	}
}

//
// fastaFilename is  either a fasta or a UM fasta (.fa or .umfa)
// filename.  In both cases, the suffix gets removed and the
// base name is kept for later use depending on context.
//
int GenomeSequence::setFastaName(String fastaFilename)
{
	if(fastaFilename.Right(3)==".fa") {
		baseFilename = fastaFilename.Left(fastaFilename.Length() - 3);
		this->fastaFilename = fastaFilename;
		return 0;
	}

	if(fastaFilename.Right(5)==".umfa") {
		baseFilename = fastaFilename.Left(fastaFilename.Length() - 5);
		this->fastaFilename = fastaFilename;
		return 0;
	}

	fprintf(stderr,"Input fasta format file '%s' needs .fa suffix.\n",
		(const char *) fastaFilename);
	return -1;
}

bool GenomeSequence::setChromosomeMD5andLength()
{
	MD5_CTX md5Context;
	uint8_t md5Signature[MD5_DIGEST_LENGTH];

	chromosomeInfo *c = &header->chromosomes[header->chromosomeCount - 1];
	c->size = header->elementCount - c->start + 1;

	//
	// it's easier to ensure we do this right if we just do it
	// in one big chunk:
	//
	char *md5Buffer = (char *) malloc(c->size);

	MD5_Init(&md5Context);

	for(genomeIndex_t i = 0; i < c->size; i ++) { md5Buffer[i] = (*this)[c->start + i]; }
	MD5_Update(&md5Context, md5Buffer, c->size);
	MD5_Final((unsigned char *) &md5Signature, &md5Context);
	free(md5Buffer);
	for(int i=0;i<MD5_DIGEST_LENGTH;i++) {
		// cheesy but works:
		sprintf(c->md5+2*i, "%02x", md5Signature[i]);
	}
	// redundant, strictly speaking due to sprintf NUL terminating
	// it's output strings, but put it here anyway.
	c->md5[2*MD5_DIGEST_LENGTH] = '\0';
	return false;
}

//
// recreate the umfa file from a reference fasta format file
//
int GenomeSequence::create(bool isColorSpace)
{
	// quick sanity check...
	assert(int2base[base2int[(int) 'A']]=='A');
	assert(int2base[base2int[(int) 'a']]=='A');
	assert(int2base[base2int[(int) 'T']]=='T');
	assert(int2base[base2int[(int) 't']]=='T');
	assert(int2base[base2int[(int) 'C']]=='C');
	assert(int2base[base2int[(int) 'c']]=='C');
	assert(int2base[base2int[(int) 'G']]=='G');
	assert(int2base[base2int[(int) 'g']]=='G');
	assert(int2base[base2int[(int) 'N']]=='N');
	assert(int2base[base2int[(int) 'n']]=='N');
	assert(int2base[base2int[(int) 'M']]=='M');
	assert(int2base[base2int[(int) 'm']]=='M');

	assert(base2int[(int) 'N']==4);
	assert(base2int[(int) 'n']==4);
	assert(base2int[(int) 'A']==0);
	assert(base2int[(int) 'a']==0);
	assert(base2int[(int) 'T']==3);
	assert(base2int[(int) 't']==3);
	assert(base2int[(int) 'C']==1);
	assert(base2int[(int) 'c']==1);
	assert(base2int[(int) 'G']==2);
	assert(base2int[(int) 'g']==2);

	if(baseFilename=="") {
		fprintf(stderr,"Base reference filename is empty.\n");
		return -1;
	}

	umfaFilename = baseFilename + ".umfa";

	if(access(umfaFilename, R_OK) == 0) {
		fprintf(stderr,"Output file '%s' exists or is not writable - please remove.\n",
			(const char *) umfaFilename);
		return -1;
	}

	MemoryMap fastaFile;

	if(fastaFile.open(fastaFilename)) {
		fprintf(stderr,"failed to open input fasta file '%s'.\n",
			(const char *) fastaFilename);
		return -1;
	}

	fprintf(stderr,"Creating FASTA binary cache file '%s'.\n", (const char *)umfaFilename);

	size_t fastaDataSize = fastaFile.length();

	size_t estimatedElementCount = fastaDataSize;

	if(genomeSequenceArray::create(umfaFilename, estimatedElementCount)) {
		fprintf(stderr,"failed to create '%s'.\n", (const char *)umfaFilename);
		perror("");
		return -1;
	}
	header->elementCount = 0;
	header->isColorSpace = isColorSpace;

	String chromosomeName;

	char *sequence = (char *) fastaFile.data;
	signed char lastBase = 5;    // for converting the reference to colorspace, the first bp is always 5 (in base space it is 'N')
	while(fastaDataSize--) {
		switch(*sequence) {
			case '\n':
			case '\r':
				sequence++;
				continue;
			case '>': {
				chromosomeName = "";
				sequence++;
				// pull out the chromosome new name
				while(!isspace(*sequence)) {
					chromosomeName += *sequence++;
				}
				// eat the rest of the line
				while(*sequence!='\n' && *sequence!='\r') {
					sequence++;
				}
				// save the Chromosome name and index into our
				// header so we can use them later.
				chromosomeInfo *c = &header->chromosomes[header->chromosomeCount];
				c->setChromosomeName(chromosomeName);
				c->start = header->elementCount;
				// c->size gets computed at the next '>' line or at the EOF

				if(header->chromosomeCount>0) {
					setChromosomeMD5andLength();
				}
				header->chromosomeCount++;
				continue;
			}
			default:
				// save the base pair value
				// Note: invalid characters come here as well, but we
				// let ::set deal with mapping them.
				if(isColorSpace) {
//
// anything outside these values represents an invalid base
// base codes: 0-> A,    1-> C,     2-> G,      3-> T
// colorspace: 0-> blue, 1-> green, 2-> oragne, 3->red
//
const char fromBase2CS[] = {
	/* 0000 */ 0,   // A->A
	/* 0001 */ 1,   // A->C
	/* 0010 */ 2,   // A->G
	/* 0011 */ 3,   // A->T
	/* 0100 */ 1,   // C->A
	/* 0101 */ 0,   // C->C
	/* 0110 */ 3,   // C->G
	/* 0111 */ 2,   // C->T
	/* 1000 */ 2,   // G->A
	/* 1001 */ 3,   // G->C
	/* 1010 */ 0,   // G->G
	/* 1011 */ 1,   // G->T
	/* 1100 */ 3,   // T->A
	/* 1101 */ 2,   // T->C
	/* 1110 */ 1,   // T->G
	/* 1111 */ 0,   // T->T
};
					//
					// we are writing color space values on transitions,
					// so we don't write a colorspace value when we
					// get the first base value.
					//
					// On second and subsequent bases, write based on
					// the index table
					//
					char thisBase = base2int[(int) (*sequence++)];
					if(lastBase>=0) {
						char color;
						if(lastBase>3 || thisBase>3) color=4;
						else color = fromBase2CS[(int) (lastBase<<2 | thisBase)];
						// re-use the int to base, because ::set expects a base char (ATCG), not
						// a color code (1234).  It should only matter on final output.
						set(header->elementCount++, int2base[(int) color]);
					}
					lastBase = thisBase;
				} else {
					set(header->elementCount++, toupper(*sequence++));
				}
				continue;
		}
	}

	if(header->chromosomeCount==0) {
		fprintf(stderr,"No chromosomes found - aborting!\n");
		fastaFile.close();
		exit(1);
	} else {
		setChromosomeMD5andLength();
	}

#if 0
	sanityCheck(fastaFile);
#endif
	fastaFile.close();

	fprintf(stderr,"FASTA binary cache file '%s' created.\n", (const char *)umfaFilename);

	//
	// leave the umfastaFile open in case caller wants to use it
	//
	return 0;
}

int GenomeSequence::getChromosomeCount()
{
	return header->chromosomeCount;
}

//return chromosome index: 0, 1, ... 24;
int GenomeSequence::getChromosome(genomeIndex_t position)
{
	if (header->chromosomeCount == 0)
		return -1;

	int start = 0;
	int stop = header->chromosomeCount - 1;

	// eliminate case where position is in the last chromosome, since the loop
	// below falls off the end of the list if it in the last one.
	
	if(position > header->chromosomes[stop].start)
		return (stop);

	while (start <= stop)
	{
		int middle = (start + stop) / 2;

		if (position >= header->chromosomes[middle].start && position < header->chromosomes[middle + 1].start)
			return middle;

		if (position == header->chromosomes[middle + 1].start)
			return (middle + 1);

		if (position > header->chromosomes[middle + 1].start)
			start = middle + 1;

		if (position < header->chromosomes[middle].start)
			stop = middle - 1;
	}

	return -1;
}

//
// Given a chromosome name and 1-based chromosome index, return the
// genome index (0 based) into sequence for it.
//
// NB: the header->chromosomes array contains zero based genome positions
//
genomeIndex_t GenomeSequence::getGenomePosition(
	const char *chromosomeName,
	unsigned int chromosomeIndex)
{
	genomeIndex_t i = getGenomePosition(chromosomeName);
	if(i == INVALID_GENOME_INDEX) return INVALID_GENOME_INDEX;
	return i + chromosomeIndex - 1;
}

genomeIndex_t GenomeSequence::getGenomePosition(
	int chromosome,
	unsigned int chromosomeIndex)
{
	genomeIndex_t i = header->chromosomes[chromosome].start;
	if(i == INVALID_GENOME_INDEX) return INVALID_GENOME_INDEX;
	return i + chromosomeIndex - 1;
}

//
// return the genome index (0 based) of the start of the named
// chromosome.  If none is found, INVALID_GENOME_INDEX is returned.
//
// XXX may need to speed this up - and smarten it up with some
// modest chromosome name parsing.... e.g. '%d/X/Y' or 'chr%d/chrX/chrY' or
// other schemes.
//
genomeIndex_t GenomeSequence::getGenomePosition( const char *chromosomeName)
{
	int chromosome = getChromosome(chromosomeName);
	if(chromosome==INVALID_CHROMOSOME_INDEX) return INVALID_GENOME_INDEX;
	return header->chromosomes[chromosome].start;
}

int GenomeSequence::getChromosome(const char *chromosomeName)
{
	unsigned int i;
		for(i=0;i<header->chromosomeCount;i++) {
			if(strcmp(header->chromosomes[i].name, chromosomeName)==0) {
			return i;
		}
	}
	return INVALID_CHROMOSOME_INDEX;
}

//
// Given a read, reverse the string and swap the base
// pairs for the reverse strand equivalents.
//
void GenomeSequence::getReverseRead(String &read)
{
	String newRead;
	for(int i=read.Length() - 1;i>=0;i--) {
		newRead.Add(BasePair(read[i]));
	}
	read = newRead;
}

#define ABS(x) ( (x) > 0 ? (x) : -(x) )
int GenomeSequence::debugPrintReadValidation(
		String &read,
		String &quality,
		char   direction,
		genomeIndex_t   readLocation,
		int sumQuality,
		int mismatchCount,
		bool recurse
	)
{
	int validateSumQ = 0;
	int validateMismatchCount = 0;
	int rc = 0;
	String genomeData;

	for(int i=0;i<read.Length();i++) {
		if(tolower(read[i]) != tolower((*this)[readLocation + i]) ) {
			validateSumQ += quality[i] - '!';
			if(direction=='F' ? i<24 : (i >= (read.Length() - 24)) ) validateMismatchCount++;
			genomeData.Add(tolower((*this)[readLocation + i]));
		} else {
			genomeData.Add(toupper((*this)[readLocation + i]));
		}
	}
	assert(validateSumQ>=0);
	if(validateSumQ != sumQuality && validateMismatchCount == mismatchCount) {
		printf("SUMQ: Original Genome: %s  test read: %s : actual sumQ = %d, test sumQ = %d\n",
			genomeData.c_str(),
			read.c_str(),
			validateSumQ,
			sumQuality
		);
		rc++;
	} else if(validateSumQ == sumQuality && validateMismatchCount != mismatchCount) {
		printf("MISM: Original Genome: %s  test read: %s : actual mismatch %d test mismatches %d\n",
			genomeData.c_str(),
			read.c_str(),
			validateMismatchCount,
			mismatchCount
		);
		rc++;
	} else if(validateSumQ != sumQuality && validateMismatchCount != mismatchCount) {
		printf("BOTH: Original Genome: %s  test read: %s : actual sumQ = %d, test sumQ = %d, actual mismatch %d test mismatches %d\n",
			genomeData.c_str(),
			read.c_str(),
			validateSumQ,
			sumQuality,
			validateMismatchCount,
			mismatchCount
		);
		rc++;
	}

	if( recurse && ABS(validateMismatchCount - mismatchCount) > read.Length()/2) {
		printf("large mismatch difference, trying reverse strand: ");
		String reverseRead = read;
		String reverseQuality = quality;
		getReverseRead(reverseRead);
		reverseQuality.Reverse();
		rc = debugPrintReadValidation(reverseRead, reverseQuality, readLocation, sumQuality, mismatchCount, false);
	}
	return rc;
}
#undef ABS


bool GenomeSequence::wordMatch(unsigned int index, String &word)
{
	for(unsigned int i = 0; (int) i<word.Length(); i++) {
		if((*this)[index + i] != word[i]) return false;
	}
	return true;
}

int GenomeSequence::printNearbyWords(unsigned int index, unsigned int deviation, String &word)
{
	for(unsigned int i = index - deviation; i < index + deviation; i++) {
		if(wordMatch(i, word)) {
			fprintf(stderr,"word '%s' found %d away from position %u.\n",
				(const char *) word,
				(int) index - i,
				index
			);
		}
	}
	return 0;
}

void GenomeSequence::dumpSequenceSAMDictionary(std::ostream &file)
{
	for(unsigned int i=0; i<header->chromosomeCount; i++) {
		file
			<<  "@SQ:"
			<< "\tSN:" << header->chromosomes[i].name  // name
			<< "\tLN:" << header->chromosomes[i].size  // number of bases
			<< "\tAS:" << header->assemblyID           // e.g. NCBI36.3
			<< "\tM5:" << header->chromosomes[i].md5
			<< "\tUR:" << header->uri
			<< "\tSP:" << header->species              // e.g. Homo_sapiens
			<< std::endl;
	}
}

void GenomeSequence::getString(String &str, genomeIndex_t index, int baseCount)
{
	str.Clear();
	if(baseCount > 0) {
		for(int i=0;i<baseCount;i++) {
			str.Add((*this)[index + i]);
		}
	} else {
		// if caller passed negative basecount, give them
		// the read for the 3' end
		for(int i=0;i< -baseCount;i++) {
			str.Add(base2complement[(int) (*this)[index + i]]);
		}
	}
}

void GenomeSequence::print30(genomeIndex_t index)
{
	std::cout << "index: " << index << "\n";
	for(genomeIndex_t i=index-30; i<index+30;i++)
		std::cout << (*this)[i];
	std::cout << "\n";
	for(genomeIndex_t i=index-30; i<index;i++)
		std::cout << " ";
	std::cout << "^";
	std::cout << std::endl;
}


#if defined(TEST)
//
// After I build libcsg, I compile and run this test code using:
//
// g++ -DTEST -o try GenomeSequence.cpp -L. -lcsg -lm -lz -lssl
// ./try
//
int main(int argc, const char **argv)
{
	GenomeSequence reference;
	String referenceName = "/usr/cluster/share/karma/NCBI36_5000.fa";
	if(reference.setFastaName(referenceName)) {
		fprintf(stderr,"failed to open reference file %s.\n", referenceName.c_str());
		exit(1);
	}
	fprintf(stderr, "open and prefetch the reference genome: ");
	// open them
	if(reference.open()) {
		exit(1);
	}
	fprintf(stderr, "done!\n");
			
	//
	// For the human genome, genomeIndex ranges from 0 to 3.2x10^9
	//
	genomeIndex_t   genomeIndex;    // 0 based
	unsigned int chromosomeIndex;   // 1 based
	unsigned int chromosome;        // 0..23 or so
	String chromosomeName;

	//
	// Here we'll start with a chromosome name, then obtain the genome
	// index, and use it to find the base we want:
	//
	chromosomeName = "2";
	chromosomeIndex = 1234567;
	// this call is slow (string search for chromsomeName):
	genomeIndex = reference.getGenomePosition(chromosomeName.c_str(), chromosomeIndex);
	assert(genomeIndex!=INVALID_GENOME_INDEX);
	printf("Chromosome %s, index %d contains base %c at genome index position %ld\n",
			chromosomeName.c_str(),
			chromosomeIndex,
			reference[genomeIndex],
			genomeIndex);

	//
	// now reverse it - given a genomeIndex from above, find the chromosome
	// name and index:
	//

	// slow (binary search on genomeIndex):
	chromosome = reference.getChromosome(genomeIndex);
	unsigned int newChromosomeIndex;
	// not slow:
	newChromosomeIndex = genomeIndex - reference.getChromosomeStart(chromosome) + 1;

	assert(chromosomeIndex == newChromosomeIndex);

	// more testing... at least test and use PackedRead:
	//
	PackedRead pr;

	pr.set("ATCGATCG", 0);
	assert(pr.size()==8);
	assert(pr[0]==GenomeSequence::base2int[(int) 'A']);
	assert(pr[1]==GenomeSequence::base2int[(int) 'T']);
	assert(pr[2]==GenomeSequence::base2int[(int) 'C']);
	assert(pr[3]==GenomeSequence::base2int[(int) 'G']);
	pr.set("ATCGATCG", 1);
	assert(pr.size()==9);
	pr.set("", 0);
	assert(pr.size()==0);
	pr.set("", 1);
	assert(pr.size()==1);
	pr.set("", 2);
	assert(pr.size()==2);
	pr.set("", 3);
	assert(pr.size()==3);
	assert(pr[0]==GenomeSequence::base2int[(int) 'N']);
	assert(pr[1]==GenomeSequence::base2int[(int) 'N']);
	assert(pr[2]==GenomeSequence::base2int[(int) 'N']);
	pr.set("C", 1);
	assert(pr.size()==2);
	assert(pr[0]==GenomeSequence::base2int[(int) 'N']);
	assert(pr[1]==GenomeSequence::base2int[(int) 'C']);

	exit(0);
}

#endif
