#ifndef _GENOME_SEQUENCE_H
#define _GENOME_SEQUENCE_H

#include <assert.h>
#include <openssl/md5.h>        // for MD5_DIGEST_LENGTH
#include <string.h>
#include "MemoryMapArray.h"
#include "StringBasics.h"

#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS
#endif
#include <stdint.h>

typedef uint32_t    genomeIndex_t;
#define INVALID_GENOME_INDEX UINT32_MAX

// chromosome index is just a signed int, so this is ok here:
#define INVALID_CHROMOSOME_INDEX -1


#define MAX_GENOME_INFO_STRING 64
#define MAX_CHROMOSOME_COUNT 100

struct chromosomeInfo {
	chromosomeInfo() {memset(this,0, sizeof(*this));}
	void setChromosomeName(const char *n) {strcpy(name, n);}
	genomeIndex_t                   start;
	genomeIndex_t                   size;
	char                            md5[2*MD5_DIGEST_LENGTH + 1];    // 32 chars plus NUL
	char                            name[MAX_GENOME_INFO_STRING];
};

class genomeSequenceMmapHeader : public MemoryMapArrayHeader {
public:
	unsigned int chromosomeCount;
	char        assemblyID[MAX_GENOME_INFO_STRING];
	char        species[MAX_GENOME_INFO_STRING];
	char        uri[MAX_GENOME_INFO_STRING];
	bool        isColorSpace;

	chromosomeInfo  chromosomes[MAX_CHROMOSOME_COUNT];
	size_t  getSize() { return sizeof(*this); }
	// return TRUE if it failed, false otherwise (primarily a length check).
	bool setAssemblyID(String &newID) {if(newID.Length()>MAX_GENOME_INFO_STRING) return true; strcpy(assemblyID, newID); return false;}
	bool setSpecies(String &newSpecies) {if(newSpecies.Length()>MAX_GENOME_INFO_STRING) return true; strcpy(species, newSpecies); return false;}
	bool setURI(String &newURI) {if(newURI.Length()>MAX_GENOME_INFO_STRING) return true; strcpy(uri, newURI); return false;}
	void setHeaderSize() {headerSize = sizeof(*this);}
};

//
// define the genomeSequence array type:
//
// NB the access/set routines use the encoded base values in the range
// 0-15, not the corresponding base pair character.
//
inline uint8_t genomeSequenceAccess(void *base, genomeIndex_t index)
{
	if((index&1)==0) {
		return ((uint8_t *) base)[index>>1] & 0xf;
	} else {
		return (((uint8_t *) base)[index>>1] & 0xf0) >> 4;
	}
};
inline void genomeSequenceSet(void *base, genomeIndex_t index, uint8_t v)
{
	if((index&1)==0) {
		((uint8_t *) base)[index>>1] = (((uint8_t *) base)[index>>1] & 0xf0) | v;
	} else {
		((uint8_t *) base)[index>>1] = (((uint8_t *) base)[index>>1] & 0x0f) | v<<4;
	}
}

inline size_t mmapGenomeSequenceElementCount2Bytes(genomeIndex_t i) {return sizeof(uint8_t) * i / 2;}

#define UMFA_COOKIE 0x1b7933a1  // unique cookie id
#define UMFA_VERSION 20090617U  // YYYYMMDD of last change to file layout

typedef MemoryMapArray<
    uint8_t,
	genomeIndex_t,
	UMFA_COOKIE,
	UMFA_VERSION,
	genomeSequenceAccess,
	genomeSequenceSet,
	mmapGenomeSequenceElementCount2Bytes, 
	genomeSequenceMmapHeader
> genomeSequenceArray;

class PackedRead {
	void set(int index, int val) {
		packedBases[index>>1] =
			(packedBases[index>>1]             // original value
			& ~(7<<((index&0x01)<<2)))         // logical AND off the original value
			| ((val&0x0f)<<((index&0x1)<<2));  // logical OR in the new value
	}
public:
	vector<uint8_t> packedBases;
	uint32_t    length;
	int size() {return length;}
	void clear() {packedBases.clear();length=0;}
	void set(const char *rhs, int padWithNCount = 0);
	uint8_t operator [] (int index) {return (packedBases[index>>1] >> ((index&0x1)<<2)) & 0xf;}
};

// String &operator = (String &lhs, const PackedRead &rhs);

class GenomeSequence : public genomeSequenceArray {
public:
	static const int baseAIndex = 000;
	static const int baseTIndex = 001;
	static const int baseCIndex = 002;
	static const int baseGIndex = 003;
	static const int baseNIndex = 004; // baseN --> bad read
	static const int baseXIndex = 005; // baseX --> unknown (bad data)
	//
	// two arrays for converting back and forth between base pair character
	// value (ASCII) to a base integer in the range 0..3.  Note there is actually
	// a value 4 and 5, for 'N' (indelible) and 'M' (unknown to me).
	//
	static unsigned char base2int[];        // base space read (ATCG)
	static const char int2base[];
	static unsigned char base2complement[];

private:
	int                             debugFlag;
	String				baseFilename;	// for later use by WordIndex create and open
	String				fastaFilename;
	String				umfaFilename;
	MemoryMap			umfaFile;

public:
	GenomeSequence();
	~GenomeSequence();

	int open(bool isColorSpace = false);
	int create(bool isColorSpace = false);
	int setFastaName(String fastaFilename);
	const String &getFastaName() {return fastaFilename;}

	bool isColorSpace() { return header->isColorSpace;}
	//
	genomeIndex_t   getNumberBases() {
		return (genomeIndex_t) header->elementCount;
	}

	int getChromosome(genomeIndex_t position);
	int getChromosome(const char *chromosomeName);
	int getChromosomeCount();

	genomeIndex_t getChromosomeStart(int chromosomeIndex) {
		return header->chromosomes[chromosomeIndex].start;
	}

	genomeIndex_t getChromosomeSize(int chromosomeIndex) {
		return header->chromosomes[chromosomeIndex].size;
	}

	//
	// get the 0 based genome index for the given
	// chromosome name and chromosome index (1 based)
	//
	genomeIndex_t getGenomePosition(
		const char *chromosomeName,
		unsigned int chromosomeIndex);

	genomeIndex_t getGenomePosition(
		int chromosome,
		unsigned int chromosomeIndex);

	// given the chromosome name, get the corresponding 0 based genome index
	// for the start of that chromosome
	genomeIndex_t getGenomePosition( const char *chromosomeName);
	genomeIndex_t getGenomePosition(int chromosomeIndex);

	const String &getBaseFilename() {return baseFilename;}
	const char *getChromosomeName(int chromosomeIndex) {
		return header->chromosomes[chromosomeIndex].name;
	}

	void setDebugFlag(bool d) {debugFlag = d;}

	unsigned int sequenceLength() { return (genomeIndex_t) header->elementCount; }
	const char *chromosomeName(int chr) {
		return header->chromosomes[chr].name;
	}

	void sanityCheck(MemoryMap &fasta);
	String IntegerToSeq (unsigned int n, unsigned int wordsize);

	bool wordMatch(unsigned int index, String &word);
	int printNearbyWords(unsigned int index, unsigned int variance, String &word);
	char BasePair(char c) {return base2complement[(int) c];}
	void dumpSequenceSAMDictionary(std::ostream&);
	inline char operator[] (genomeIndex_t index) {
		return int2base[(*((genomeSequenceArray*) this))[index]];
	}
	inline uint8_t getInteger(genomeIndex_t index) {
		return (*((genomeSequenceArray*) this))[index];
	}
	inline void set(genomeIndex_t index, char value) {
		genomeSequenceArray::set(index, base2int[(uint8_t) value]);
	}
	//
	// this is a fairly ugly hack to reach into the 
	// raw genome vector, get the byte that encodes
	// two bases, and return it.
	uint8_t *getDataPtr(genomeIndex_t index) {
		return ((uint8_t *) data + index/2);
	}
	//
	// when creating the genome mapped file, we call this to set
	// the MD5 checksum of the chromosome sequence and length so that
	// we can write the SAM SQ headers properly.
	// NB: operates on the last fully loaded chromosome.
	bool setChromosomeMD5andLength();
	bool setAssemblyID(String &newID) {return header->setAssemblyID(newID);}
	bool setSpecies(String &newSpecies) {return header->setSpecies(newSpecies);}
	bool setURI(String &newURI) {return header->setURI(newURI);}

	// replace read with the reversed one
	void getReverseRead(String &read);
	// debug the given read - print nice results
	int debugPrintReadValidation(
			String &read,
			String &quality,
			char   direction,
			genomeIndex_t   readLocation,
			int sumQuality,
			int mismatchCount,
			bool recurse = true
	);

	void getString(String &str, genomeIndex_t index, int baseCount);

	void print30(genomeIndex_t);
};


#endif
