#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>

#include "MemoryMap.h"

MemoryMap::MemoryMap() {
	constructor_clear();
	page_size = sysconf(_SC_PAGE_SIZE);
};

MemoryMap::~MemoryMap() {
	destructor_clear();
};

void MemoryMap::debug_print() {
	printf("fd = %d\n", fd);
	printf("data = 0x%p\n", data);
	printf("offset = 0x%lx\n", offset);
	printf("mapped_length = 0x%lx\n", mapped_length);
	printf("total_length = 0x%lx\n", total_length);
	printf("page_size = 0x%lx\n", page_size);
};

void MemoryMap::constructor_clear() {
	fd = -1;
	data = (void *) NULL;
	offset = 0;
	mapped_length = 0;
	total_length = 0;
	useMemoryMapFlag = true;
};

void MemoryMap::destructor_clear() {
	if(data!=NULL) {
		munmap(data, mapped_length);
	}
	if (fd!=-1) {
		::close(fd);
	}
	constructor_clear();
};

int MemoryMap::open(String file, int flags) {

	struct stat buf;

	int mmap_prot_flag = PROT_READ;
	if (flags != O_RDONLY) mmap_prot_flag = PROT_WRITE;

	fd = ::open((const char *) file, flags);
	if (fd==-1) {
		fprintf(stderr, "MemoryMap::open: file %s not found\n", (const char *) file);
		constructor_clear();
		return -1;
	}
	if(fstat(fd, &buf)) {
		perror("MemoryMap::open");
		constructor_clear();
		return -1;
	}
	mapped_length = buf.st_size;
	total_length = mapped_length;

	if(useMemoryMapFlag) {

		data = ::mmap(
				NULL,           // start
				mapped_length,
				mmap_prot_flag, // protection flags
				MAP_SHARED | MAP_POPULATE | MAP_NONBLOCK, // share/execute/etc flags
				fd,
				offset
		);
		if(data == MAP_FAILED) {
			::close(fd);
			perror("MemoryMap::open");
			constructor_clear();
			return -1;
		}
	} else {
		data = (void *) malloc(mapped_length);
		if(data==NULL) {
			::close(fd);
			perror("MemoryMap::open");
			constructor_clear();
			return -1;
		}
		ssize_t resultSize = read(fd, data, mapped_length);
		if(resultSize!=mapped_length) {
			::close(fd);
			perror("MemoryMap::open");
			constructor_clear();
			return -1;
		}
	}
	return 0;
};

int MemoryMap::create(String file, size_t size) {

	int mmap_prot_flag = PROT_READ | PROT_WRITE;

	fd = ::open((const char *) file, O_RDWR|O_CREAT|O_TRUNC, 0666);
	if (fd==-1) {
		fprintf(stderr, "MemoryMap::open: can't create file '%s'\n",(const char *) file);
		constructor_clear();
		return -1;
	}

	lseek(fd, (off_t) size - 1, SEEK_SET);
	char ch = 0;
	write(fd, &ch, 1);

	data = ::mmap(
			NULL,           // start
			size,
			mmap_prot_flag, // protection flags
			MAP_SHARED,     // share/execute/etc flags
			fd,
			offset
	);
	if(data == MAP_FAILED) {
		::close(fd);
		unlink(file);
		perror("MemoryMap::open");
		constructor_clear();
		return -1;
	}
	mapped_length = size;
	total_length = size;
	return 0;
};

int MemoryMap::close() {
	destructor_clear();
	return 0;
};

void MemoryMap::test()
{
	int result;

	result = this->open("test/test_memmap_data.txt");
	assert(result == 0);
	assert(data!=NULL);
	assert(mapped_length == 183);   // length of the above file
	close();

	// now try non memory mapped (direct slow file I/O)
	useMemoryMap(false);
	result = this->open("test/test_memmap_data.txt");
	assert(result == 0);
	assert(data!=NULL);
	assert(mapped_length == 183);   // length of the above file
	close();
}

int MemoryMap::prefetch()
{
	int sum = 0;
	size_t i;
	size_t pageSize = getpagesize();

	for(i=0;i<mapped_length;i += pageSize) sum += *(i + (char *) data);

	return sum;
}

#if defined(TEST)
//
// compile test using:
// g++ -DTEST -o testMemoryMap MemoryMap.cpp StringBasics.o Error.o -lz
//

int main(int argc, const char *argv)
{
	MemoryMap map;

	map.test();

//	map.debug_print();

	exit(0);
}
#endif

