#ifndef __MEMORYMAPARRAY_H
#define __MEMORYMAPARRAY_H

#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS
#endif
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <time.h>

#include "MemoryMap.h"


//
// This file defines a template for generating memory map backed arrays
// of different types of values.
//
// The template assumes that the mapped files are broken into two parts,
// first, a header (MemoryMapArrayHeader), then followed by the data
// in the array.
//
// typedefs are used to declare various types of arrays beforehand,
// since there will be only a few.
//
// They are:
//      mmapArrayUint32_t;
//      mmapArrayBool_t;
//      mmapArray4Bit_t;
//
// XXX consider adding env("USER"), argv[0], date/time creation, etc.
//
class MemoryMapArrayHeader {
public:
	void constructorClear() {memset(this, 0, sizeof(*this));}
	uint32_t     typeCookie;
	uint32_t     typeVersion;
	uint32_t     contentCookie;
	uint32_t     contentVersion;
	size_t      headerSize;

	// file generation info
	time_t      creationDate;
	char        creationUser[32];
	char        creationHost[32];
	char        application[32];
	// now describe the data:
	size_t      elementCount;
	void debugPrint(FILE *);
	void setHeaderSize() {headerSize = sizeof(*this);}
};

//
// This class object represents the application specific information that doesn't
// fit in the general header above.  Since it is only allocated via an mmap operation,
// as part of the mapped file, the destructor must never be called.  The virtual
// destructor is declared to eliminate gcc warnings.
//
// For many arrays, this will be empty.
//
struct MemoryMapGenericHeader {
protected:
	size_t  headerSize;
public:
	size_t getHeaderSize() {return headerSize;}
	// other stuff follows...
};

template <
	class elementT,
	typename indexT,
	unsigned int cookieVal,
	unsigned int versionVal,
	elementT accessorFunc(void *base, indexT),
	void setterFunc(void *base, indexT, elementT),
	size_t elementCount2BytesFunc(indexT),
	class arrayHeaderClass>
class MemoryMapArray : public MemoryMap
{
protected:
	arrayHeaderClass    *header;
	void                *data;
	const char          *errorStr;
public:
	void constructorClear() {
		header = NULL;
		data = NULL;
		errorStr = NULL;
	}
	MemoryMapArray() {
		constructorClear();
	}
	~MemoryMapArray() {
		if(data) close();
	}

	void setContentCookie(uint32_t c) {header->contentCookie = c;}
	void setContentVersion(uint32_t v) {header->contentVersion = v;}
	void setApplication(const char *s) {strcpy(header->application, s);}

	// accessing
	inline elementT operator[] (indexT i) {return accessorFunc(data, i);}
	inline void set(indexT i, elementT v) {setterFunc(data, i, v);}

	//
	// do administrative setup of the header and populating this
	// class members.  User will need to finish populating the
	// contents of the metaData and data sections.
	//
	int create(String file, indexT elementCount) {
		size_t len = elementCount2BytesFunc(elementCount) + sizeof(*header);
		int rc;
		rc = MemoryMap::create(file, len);
		if(rc) {
			errorStr = "failed to create file";
			return rc;
		}
		header = (arrayHeaderClass *) MemoryMap::data;
		header->constructorClear();
		header->typeCookie = cookieVal;
		header->typeVersion = versionVal;
		header->setHeaderSize();
		header->elementCount = elementCount;
		data = (elementT *) ((char *) MemoryMap::data + header->headerSize);

		const char *env;
		env = getenv("USER");
		if(env) strcpy(header->creationUser, env);
		header->creationDate = time(NULL);
		gethostname(header->creationHost, sizeof(header->creationHost));
		return 0;
	}

	int open(String file) {
		int rc = MemoryMap::open(file);
		if(rc) {
			errorStr = "failed to map file";
			return rc;
		}
		header = (arrayHeaderClass *) MemoryMap::data;
		data = (elementT *) ((char *) MemoryMap::data + header->headerSize);
		if(header->typeCookie!=cookieVal) {
			close();
			errorStr = "got bad cookie";
			// XXX insert better error handling
			return 1;
		}
		if(header->typeVersion!=versionVal) {
			errorStr = "got bad version";
			close();
			// XXX insert better error handling
			return 1;
		}
		return 0;
	}

	void close() {
		constructorClear();
		MemoryMap::close();
	}
	void debugPrint(FILE *f) {if(header) header->debugPrint(f);}
};

struct emptyGenericHeader : public MemoryMapGenericHeader {
public:
	void setHeaderSize() {headerSize = sizeof(*this);}
};

//
// define the uint32 array type:
//
inline uint32_t mmapUint32Access(void *base, uint32_t index) {return ((uint32_t *)base)[index];}
inline void mmapUint32Set(void *base, uint32_t index, uint32_t v) {((uint32_t *)base)[index] = v;}
inline size_t mmapUint32elementCount2Bytes(uint32_t i) {return sizeof(uint32_t) * i;}

typedef MemoryMapArray<
	uint32_t,
	uint32_t,
	0x16b3816c,
	20090109,
	mmapUint32Access,
	mmapUint32Set,
	mmapUint32elementCount2Bytes,
	MemoryMapArrayHeader
	> mmapArrayUint32_t;

//
// define the boolean memory mapped array type.
// NB: it is limited to 2**32 elements
//
inline bool mmapBoolAccess(void *base, uint32_t i) {return (((char*)base)[i>>3] >> (i&0x7)) & 0x1;}
inline void mmapBoolSet(void *base, uint32_t i, bool v)
{
	((unsigned char*) base)[i>>3] = 
		(((unsigned char*) base)[i>>3] & ~(1<<(i&0x7))) |
		((v&1)<<(i&0x7));
}
inline size_t mmapBoolelementCount2Bytes(uint32_t i) {return (size_t) (i+7)/8;}

typedef MemoryMapArray<
	bool,
	uint32_t,
	0xac6c1dc7,
	20090109,
	mmapBoolAccess,
	mmapBoolSet,
	mmapBoolelementCount2Bytes,
	MemoryMapArrayHeader
	> mmapArrayBool_t;

//
// define the two bit memory mapped array type:
//
inline uint32_t mmap2BitAccess(void *base, uint32_t i) {return (((unsigned char*)base)[i>>2] >> ((i&0x3)<<1)) & 0x3;}
inline void mmap2BitSet(void *base, uint32_t i, uint32_t v)
{
	((unsigned char*) base)[i>>2] =
		(((unsigned char*) base)[i>>2]      // original value
		& ~(3<<((i&0x03)<<1)))              // logical AND off the original value
		| ((v&0x03)<<((i&0x3)<<1));         // logical OR in the new value
}
inline size_t mmap2BitElementCount2Bytes(uint32_t i) {return (size_t) (i+3)/4;}

typedef MemoryMapArray<
	uint32_t,
	uint32_t,
	0x25b3ea5f,
	20090109,
	mmap2BitAccess,
	mmap2BitSet,
	mmap2BitElementCount2Bytes,
	MemoryMapArrayHeader
	> mmapArray2Bit_t;

//
// define the four bit memory mapped array type:
//
inline uint32_t mmap4BitAccess(void *base, uint32_t i) {return (((unsigned char*)base)[i>>1] >> ((i&0x1)<<2)) & 0xf;}
inline void mmap4BitSet(void *base, uint32_t i, uint32_t v)
{
	((unsigned char*) base)[i>>1] =
		(((unsigned char*) base)[i>>1]      // original value
		& ~(7<<((i&0x01)<<2)))              // logical AND off the original value
		| ((v&0x0f)<<((i&0x1)<<2));         // logical OR in the new value
}
inline size_t mmap4BitelementCount2Bytes(uint32_t i) {return (size_t) (i+1)/2;}

typedef MemoryMapArray<
	uint32_t,
	uint32_t,
	0x418e1874,
	20090109,
	mmap4BitAccess,
	mmap4BitSet,
	mmap4BitelementCount2Bytes,
	MemoryMapArrayHeader
	> mmapArray4Bit_t;

#if 0
	// XXX this is example code I want to use to define arrays of genome wide match values
class   baseRecord {
	unsigned int base:4;
	unsigned int qScore:7;
	unsigned int conflicts:5;   // how many cases of poorer matches that disagree
};

//
// define the baseRecord array type:
//
inline baseRecord &mmapBaseRecordAccess(void *base, uint32_t index) {return *((baseRecord *) ((char *)base + index*sizeof(baseRecord)));}
inline void mmapBaseRecordSet(void *base, uint32_t index, baseRecord &v) {mmapBaseRecordAccess(base, index) = v;}
inline size_t mmapBaseRecordElementCount2Bytes(uint32_t i) {return sizeof(baseRecord) * i;}

typedef MemoryMapArray<
	baseRecord &,
	uint32_t,
	0x12341234,
	0xdeadbeef,
	&mmapBaseRecordAccess,
	mmapBaseRecordSet,
	mmapBaseRecordElementCount2Bytes,
	MemoryMapArrayHeader
> mmapArrayBaseRecord_t;
#endif

#endif
