
#include <stdio.h>
#include <stdlib.h>
#include "SmithWaterman.h"

// put TEST below here, so that makedepend will see the .h, so that we
// can get a clean dependency for SmithWaterman.o, so that we can at least
// compile the header when we change it.

#if defined(TEST)
// g++ -g -o testSW -DTEST SmithWaterman.cpp
//
// Smith-Waterman - test code uses a 256x256 array of int16
//
int swat(const char *A,
	const char *qualities,
	const char *B,
	int direction,
	const char *expectedCigarString,
	int expectedSumQ
	)
{
	int allowedInsertDelete = 1;
	int errors = 0;

	SmithWaterman<uint16_t [256][256], uint16_t, const char *, const char *, const char *> sw(&A, &qualities, &B, strlen(A), strlen(B), allowedInsertDelete, direction);

	cout << "=============" << endl;
	cout << "        Read: " << A << endl;
	cout << "   Reference: " << B << endl;
	cout << "   Direction: " << direction << endl;

	sw.populateH();
	sw.printH();
	sw.populateAlignment();

	// in Karma, we'll first do a whole bunch of getSumQ calls
	int sumQ = sw.getSumQ();

	// then for the winning match, we'll get two different 
	CigarRoller cigar;
	cigar.clear();
	sw.rollCigar(cigar);
//	sw2.rollCigar(cigar);

	const char *cigarStr = cigar.getString();
	cout << "Cigar String: " << cigarStr << endl;
	if( strcmp(cigarStr, expectedCigarString)!=0) {
		cout << "ERROR: Expected Cigar String: " << expectedCigarString << endl;
		errors ++;
	}

	for(vector<pair<int,int> >::iterator i = sw.alignment.begin(); i != sw.alignment.end();i++) cout << *i << endl;
	cout << "sumQ = " << sumQ << endl;
	if( sumQ != expectedSumQ) {
		cout << "ERROR: Expected SumQ: " << expectedSumQ << endl;
		errors ++;
	}

	cout << "Max Cell: " << sw.maxCostPosition << endl;
	cout << "=============" << endl << endl;
	delete cigarStr;
	// now you can reconstruct the strings.
	return errors;
}


// test with Sequence 1 = ACACACTA
//           Sequence 2 = AGCACACA
int main(int argc, const char **argv)
{
	CigarRoller roller;
	int errors = 0;

	roller.test();

	// CIGAR explanation - for backward SW runs, the corresponding
	// CIGAR string is generated from the back of the string to the
	// front.  Recall that the soft clipping is only done at the
	// "end" of the string, taking direction into account.

	// forwards - simple
	errors += swat("1234", "\"#$-", "1235", 1, "3M1S", 0);

	// backwards - simple
	errors += swat("1234", "\"#$-", "1235", -1, "4M", 12);

	// backwards - soft left clip
	errors += swat("1234", "\"#$-", "0234", -1, "1S3M", 0);

	// delete in read (arg 1) - forward
	errors += swat("123467890", "\"#$%^&*()-", "1234567890", +1, "4M1D5M", 50);

	// insert in read (arg 1) - forward
	errors += swat("1234556789", "\"#$%^&*()-", "1234567890", +1, "5M1I4M", 50);

	// delete in read (arg 1) - backward
	errors += swat("123467890", "\"#$%^&*()-", "1234567890", -1, "4M1D5M", 50);

	// insert in read (arg 1) - backward
	errors += swat("1234556789", "\"#$%^&*()-", "0123456789", -1, "4M1I5M", 50);

	// forward - soft right clip of 2 bases - sumQ should be 0
	errors += swat("123456700", "\"#$%^&*()-", "123456789", +1, "7M2S", 0);

	// insert in read (arg 1) - forward w/2 mismatches at end
	errors += swat("1023456700", "\"#$%^&*()-", "123456789", +1, "1M1I6M2S", 50);

	cout << endl << "Total Errors found: " << errors << endl;
}
#endif
