#include "MathBinNormal.h"
#include "MathStats.h"
#include "MathConstant.h"

#include <math.h>

// Discrete Data Normal Equations class
//

DiscreteDataNormalEquations::DiscreteDataNormalEquations()
   : integrand(this), hyperCube("HyperCube"), miser()
   {
   miser.vfunc = &integrand;
   }

bool DiscreteDataNormalEquations::isUnivariate(Matrix & correl)
   {
   for (int i = 0; i < correl.rows; i++)
      for (int j = i + 1; j < correl.rows; j++)
         if (fabs(correl[i][j]) > EPS)
            return false;
   return true;
   }

double DiscreteDataNormalEquations::Evaluate()
   {
   // The variance matrix A (varMatrix)
   // and its cholesky decomposition are updated
   CalculateCovariances();

   // Work out the expected scores using linear model
   residuals.Product(linearModel, means);

   // Can we do univariate integration to save time?
   if (isUnivariate(varMatrix) || preliminary)
      {
      likelihood = 0.0;

      for (int i = 0; i < residuals.dim; i++)
         {
         double z = residuals[i] / sqrt(varMatrix[i][i]);
         likelihood += log (ndist(z, scores[i] == 0.0));
         }

      return likelihood *= multiple;
      }

   // Numerical integration carried out using Gentz's transformation
   // First we sort the integration ranges
   SortIntegrationRanges();

   // Calculate the cholesky decomposition
   cholesky.FastDecompose(varMatrix);

   // Calculate f[1] ...
   integrand.f1 = ndist(residuals[0]/cholesky.L[0][0], upper[0]);

   // Integrate numerically by Monte-Carlo

   miser.ncall = 50000 * scores.dim * scores.dim;
   likelihood = log(miser.Integrate(hyperCube));

   return likelihood *= multiple;
   }

void DiscreteDataNormalEquations::Prepare()
   {
   NormalEquations::Prepare();

   hyperCube.Dimension(2, scores.dim - 1);
   hyperCube[0].Set(0.0);
   hyperCube[1].Set(1.0);

   integrand.y.Dimension(scores.dim - 1);
   upper.Dimension(scores.dim + 1);
   }

void DiscreteDataNormalEquations::SelectPoint(Vector & v)
   {
   for (int i = 0; i < means.dim; i++)
      means[i] = v[i];

   if (preliminary)
      {
      variances.Set(1.0 / variances.dim);
      return;
      }

   double residualVar = 1.0;
   for (int i = 1; i < variances.dim; i++)
      {
      variances[i] = exp(-exp(v[i + means.dim - 1])) * residualVar;
      residualVar -= variances[i];
      }
   variances[0] = residualVar;
   }

void DiscreteDataNormalEquations::RetrievePoint(Vector & v)
   {
   for (int i = 0; i < means.dim; i++)
      v[i] = means[i];

   if (preliminary) return;

   double residualVar = 1.0;

   for (int i = 1; i < variances.dim; i++)
      {
      v[i + means.dim - 1] = log(-log(variances[i] / residualVar));
      residualVar -= variances[i];
      }
   }

void DiscreteDataNormalEquations::SortIntegrationRanges()
   {
   // Use a bubble sort to place the largest integrands last
   // This may cause trouble for problems with high dimensionality

   for (int i = 0; i < scores.dim; i++)
      upper[i] = scores[i] == 0.0;

   // Although in general only the upper triangle is used
   // It is much easier to rearrange covariances with full matrix
   for (int i = 0; i < varMatrix.rows; i++)
      for (int j = i + 1; j < varMatrix.rows; j++)
         varMatrix[j][i] = varMatrix[i][j];

   for (int i = 0; i < scores.dim - 1; i++)
      {
      int ismall = 0;
      double icmp = upper[i] ? -residuals[i] : residuals[i];

      for (int j = i + 1; j < scores.dim; j++)
         if (icmp > (upper[j] ? -residuals[j] : residuals[j]))
            ismall = j;

      if (!ismall) continue;

      varMatrix.SwapRows(i, ismall);
      varMatrix.SwapColumns(i, ismall);
      upper.Swap(i, ismall);
      residuals.Swap(i, ismall);
      }
   }

// DiscreteIntegrandFunc
//

double DiscreteIntegrandFunc::Evaluate(Vector & w)
   {
   double lastf = f1, result = f1;

   for (int i = 1; i < normal->residuals.dim; i++)
      {
      y[i - 1] = ninv(w[i-1] * lastf + (normal->upper[i-1] ? 1.0-lastf : 0.0));

      double z = normal->residuals[i];

      for (int j = 0; j < i; j++)
         z -= normal->cholesky.L[i][j] * y[j];
      z /= normal->cholesky.L[i][i];

      lastf = ndist(z, normal->upper[i] );
      result *= lastf;
      }

   return result;
   }

// DiscreteDataNormalSet
//

DiscreteDataNormalSet::DiscreteDataNormalSet() : NormalSet()
   {
   iterative = true;
   }

void DiscreteDataNormalSet::AllocateSets()
   {
   for (int i = 0; i < size; i++)
      sets[i] = new DiscreteDataNormalEquations;
  }

int DiscreteDataNormalSet::CountParameters()
   {
   return means.dim + (preliminary ? 0 : variances.dim - 1);
   }

void DiscreteDataNormalSet::GetStartingPoint(Vector & startPoint)
   {
   if (preliminary) sets[0]->means.Set(0.0);
   sets[0]->variances.Set(1.0 / sets[0]->variances.dim);

   sets[0]->RetrievePoint(startPoint);
   }


